package cs108;

import javafx.application.Application;
import javafx.geometry.VPos;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Text;
import javafx.stage.Stage;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import static javafx.beans.binding.Bindings.when;

public final class Main extends Application {
    private final static int PIECE_SIDE = 50;
    private final static int INTER_PIECE_SPACE = 3;

    public static void main(String[] args) {
        launch(args);
    }

    @Override
    public void start(Stage primaryStage) {
        PuzzleBean puzzleBean = new PuzzleBean();

        Map<Piece, Node> pieceNode = new HashMap<>();
        for (Piece piece : Piece.values()) {
            if (piece == Piece.HOLE) continue;

            Node node = nodeForPiece(piece);
            pieceNode.put(piece, node);

            node.setOnMouseClicked(e -> {
                Puzzle puzzle = puzzleBean.getPuzzle();
                int pIndex = puzzle.pieces().indexOf(piece);
                if (puzzle.canSwap(pIndex))
                    puzzleBean.setPuzzle(puzzle.withPieceSwapped(pIndex));
            });
        }

        Text motivator = new Text();
        motivator.textProperty().bind(
                when(puzzleBean.isSolvedProperty())
                        .then("Bravo!")
                        .otherwise("à résoudre..."));

        Pane gamePane = new Pane();
        gamePane.getChildren().setAll(pieceNode.values());
        double prefSide = PIECE_SIDE * Puzzle.SIDE_LENGTH + INTER_PIECE_SPACE * (Puzzle.SIDE_LENGTH + 1);
        gamePane.setPrefSize(prefSide, prefSide);

        Button shuffleButton = new Button("Mélanger !");
        shuffleButton.setOnAction(e ->
                puzzleBean.setPuzzle(puzzleBean.getPuzzle().shuffled()));

        BorderPane mainPane = new BorderPane(gamePane, motivator, null, shuffleButton, null);

        puzzleBean.puzzleProperty().addListener((p, o, n) ->
                layoutPuzzle(pieceNode, n));
        puzzleBean.setPuzzle(Puzzle.SOLVED);

        primaryStage.setScene(new Scene(mainPane));
        primaryStage.setTitle("Taquin");
        primaryStage.show();
    }

    private void layoutPuzzle(Map<Piece, Node> pieceNode, Puzzle puzzle) {
        Iterator<Piece> pieceIt = puzzle.pieces().iterator();
        for (int r = 0; r < Puzzle.SIDE_LENGTH; ++r) {
            int y = INTER_PIECE_SPACE + r * (PIECE_SIDE + INTER_PIECE_SPACE);
            for (int c = 0; c < Puzzle.SIDE_LENGTH; ++c) {
                int x = INTER_PIECE_SPACE + c * (PIECE_SIDE + INTER_PIECE_SPACE);

                Node n = pieceNode.get(pieceIt.next());
                if (n == null) continue;

                n.relocate(x, y);
            }
        }
    }

    private Node nodeForPiece(Piece piece) {
        Rectangle r = new Rectangle(PIECE_SIDE, PIECE_SIDE);
        r.setFill(Color.SALMON);
        Text t = new Text(5, 5, piece.toString());
        t.setTextOrigin(VPos.TOP);
        return new Group(r, t);
    }
}
