package cs108;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public final class Puzzle {
    public static final Puzzle SOLVED = new Puzzle(List.of(Piece.values()));
    public static final int SIDE_LENGTH = 3;

    private final List<Piece> pieces;

    public Puzzle(List<Piece> pieces) {
        this.pieces = List.copyOf(pieces);
    }

    private static boolean isValidPos(int x, int y) {
        return (0 <= x && x < SIDE_LENGTH) && (0 <= y && y < SIDE_LENGTH);
    }

    private static int posToIndex(int x, int y) {
        assert isValidPos(x, y);
        return y * SIDE_LENGTH + x;
    }

    public List<Piece> pieces() {
        return pieces;
    }

    public boolean isSolved() {
        return pieces.equals(SOLVED.pieces);
    }

    public Puzzle shuffled() {
        List<Piece> newPieces = new ArrayList<>(pieces);
        Collections.shuffle(newPieces);
        return new Puzzle(newPieces);
    }

    public boolean canSwap(int index) {
        return indexOfNeighboringHole(index) != -1;
    }

    public Puzzle withPieceSwapped(int index) {
        assert canSwap(index);
        int holeIndex = indexOfNeighboringHole(index);
        List<Piece> newPieces = new ArrayList<>(pieces);
        Collections.swap(newPieces, index, holeIndex);
        return new Puzzle(newPieces);
    }

    private boolean isHoleAt(int x, int y) {
        return isValidPos(x, y) && pieces.get(posToIndex(x, y)) == Piece.HOLE;
    }

    private int indexOfNeighboringHole(int index) {
        int x = index % SIDE_LENGTH;
        int y = index / SIDE_LENGTH;

        if (isHoleAt(x + 1, y))
            return posToIndex(x + 1, y);
        if (isHoleAt(x - 1, y))
            return posToIndex(x - 1, y);
        if (isHoleAt(x, y + 1))
            return posToIndex(x, y + 1);
        if (isHoleAt(x, y - 1))
            return posToIndex(x, y - 1);
        return -1;
    }
}
